/***************************************************************************/
/*                                                                         */
/* File: main.cpp                         Date: 07-08-2004                 */
/*                                                                         */
/* Auth: Ben Kenwright                    Email: bkenwright@xbdev.net      */
/*                                                                         */
/* Desc: Creating a torus shape (donut) using maths and opengl!            */
/*                                                                         */
/***************************************************************************/ 
/*
   Torus Tutorial
   Its one of those shapes you see all over the net, and is used for all
   sorts of demos, like mirror effects and pixel lighting.  As its easy to
   create and it is one of those shapes that looks cool :)

   You would be supprised to know, that the code that actually generates
   our donut, is only 20 lines long...'torus(...)' - amazing eh?

   Of course its only a basic shape...no normals or texture coordinates
   have been added...guess they can be added later once you understand
   how you generate a cool shape like this, and not have to load one in from
   file.

*/
/***************************************************************************/ 


/***************************************************************************/
/*                                                                         */
/* Include precompiled libraries.                                          */
/*                                                                         */
/***************************************************************************/

#pragma comment(lib, "opengl32.lib")
#pragma comment(lib, "glu32.lib")


// Windows headers
#include <windows.h>

// OpenGL main definitions, e.g. glBegin, glEnd etc
#include <gl/gl.h>

// gluLookAt(..) and gluPerspective(..)
#include <gl/glu.h>

// cos(..) and sin(..)
#include <math.h>


/***************************************************************************/
/*                                                                         */
/* Defines.                                                                */
/*                                                                         */
/***************************************************************************/

// Speaks for itself
#define X_PI       3.14159265358979323846

// Our windows width and height
#define	SCREEN_W	620
#define	SCREEN_H	620

/***************************************************************************/
/*                                                                         */
/* Globals                                                                 */
/*                                                                         */
/***************************************************************************/

// So we can switch between wire frame render and solid render
unsigned char g_bRender = 0;  


/***************************************************************************/
/*                                                                         */
/* Function Declarations                                                   */
/*                                                                         */
/***************************************************************************/

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
void StartOpenGL(HWND hWnd, HDC * hDC, HGLRC * hRC);
void EndOpenGL(HWND hWnd, HDC hDC, HGLRC hRC);
void RenderLoop(HDC hDC, HWND hWnd);


/***************************************************************************/
/*                                                                         */
/* Name: torus(..)                                                         */
/* Date: 07-08-2004                                                        */
/* Orig: bkenwright@xbdev.net                                              */
/*                                                                         */
/* Desc: Generates and draws a torus shape in opengl.  I've tried to with  */
/* a terrible asci sketch below to show how you effect the number of turns */
/* about the radius...and the number of turns about the column that makes  */
/* up the tube of the donut.                                               */
/*                                                                         */
/***************************************************************************/
/*
       Front View             Size View
         _ _ _ _                -----                 
       /        \               | | |             
     /            \             | | |              
   /                \           -----                 
  |                  |          | | |                 
  |                  |          | | |                 
  |                  |          | | |                  
  |                  |          -----                    
   \                /           | | |                 
     \            /             | | |           
       \ _ _ _ _/               -----            
                                                  
        numt=8                  numc=3  

*/
/***************************************************************************/


static void torus(int numc, int numt)
{
	float colr = 0.0f;                 // This is so our shape has some random
                                       // colour and looks better in solid draw.
    int i, j, k;                       // For loop variables.
    double s, t, x, y, z, twopi;

    twopi = 2 * (double)X_PI;
    for (i = 0; i < numc; i++) 
    {
        if(g_bRender)				    // What We Want To Render
            glBegin( GL_QUAD_STRIP );	// Render Polygons
        else 
            glBegin( GL_LINE_STRIP );   // Render Lines Instead

        for (j = 0; j <= numt; j++) 
        {
            for (k = 1; k >= 0; k--) 
            {
                s = (i + k) % numc + 0.5;
                t = j % numt;
                // Alter the 0.3 to larger values to get larger
				// or smaller tube of your donut :)
                x = (1+.3*cos(s*twopi/numc))*cos(t*twopi/numt);
                y = (1+.3*cos(s*twopi/numc))*sin(t*twopi/numt);
                z = .3 * sin(s * twopi / numc);
                glVertex3f((GLfloat)x, (GLfloat)y, (GLfloat)z);
                glColor3f( colr, 0, 1.0 );
                colr += 0.5f;
                if( colr > 1.0f ) colr = 0.0f;
            }//End for k
        }//End for j
        glEnd();
    }//End for i
}//End torus(..)



/***************************************************************************/
/*                                                                         */
/* Name: WinMain(..)                                                       */
/* Date: 07-08-2004                                                        */
/* Orig: bkenwright@xbdev.net                                              */
/*                                                                         */
/* Desc: Program Entry Point for Windows (Win32)                           */
/*                                                                         */
/*                                                                         */
/*-----------------------------> WINMAIN <---------------------------------*/
/*                                                                         */
/*         *******  *********     ****      *********    *********         */
/*         *******  *********     ****      **********   *********         */
/*         **          ***       ******     ****   ***      ***            */
/*         ****        ***      ***  ***    ****  ****      ***            */
/*           *****     ***     ***    ***   *********       ***            */
/*             ***     ***     **********   *********       ***            */
/*         *******     ***     ***    ***   ***   ****      ***            */
/*         *******     ***     ***    ***   ***    ****     ***            */
/*                                                                         */
/***************************************************************************/

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, 
				   LPSTR lpCmdLine, int iCmdShow)
{
	WNDCLASS wc;
	HWND hWnd;
	HDC hDC;
	HGLRC hRC;
	MSG msg;
	BOOL quit = FALSE;
	
	
	// register window class
	wc.style = CS_OWNDC;
	wc.lpfnWndProc = WndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = LoadIcon( NULL, IDI_APPLICATION );
	wc.hCursor = LoadCursor( NULL, IDC_ARROW );
	wc.hbrBackground = (HBRUSH)GetStockObject( BLACK_BRUSH );
	wc.lpszMenuName = NULL;
	wc.lpszClassName = "xbdev demo";
	RegisterClass( &wc );
	
	// create main window
	hWnd = CreateWindow( 
		"xbdev demo", "OpenGL Torus Demo - www.xbev.net", 
		WS_CAPTION | WS_POPUPWINDOW | WS_VISIBLE,
		0, 0, SCREEN_W, SCREEN_H,
		NULL, NULL, hInstance, NULL );
	
	// Start our OpenGL up for the window
	StartOpenGL( hWnd, &hDC, &hRC );
	
	// program main loop
	while ( !quit )
	{
		// check for messages
		if ( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE )  )
		{
			// handle or dispatch messages
			if ( msg.message == WM_QUIT ) 
			{
				quit = TRUE;
			} 
			else 
			{
				TranslateMessage( &msg );
				DispatchMessage( &msg );
			}
		} 
		else // No messages are in the loop, hence do rendering work
		    RenderLoop(hDC, hWnd);

	}//End while(..)

	// shutdown OpenGL, as we've finished with it and are about to exit
	EndOpenGL( hWnd, hDC, hRC );
	
	// destroy the window explicitly
	DestroyWindow( hWnd );
	
	return (int)(msg.wParam);
	
}// End WinMain(..)


/***************************************************************************/
/*                                                                         */
/* Name: WndProc(..)                                                       */
/* Date: 07-08-2004                                                        */
/* Orig: bkenwright@xbdev.net                                              */
/*                                                                         */
/* Desc: Windows Callback function - this is what we pass as one of our    */
/* windows parameters when we create a new window.  And is called whenever */
/* something with our window happens....for example, if the mouse moves    */
/* over the window...if we click something..it even gets called when we    */
/* first create or window, and when its about to die.                      */
/*                                                                         */
/***************************************************************************/

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message)
    {
        case WM_CLOSE:
		    PostQuitMessage( 0 );
		return 0;
            case WM_DESTROY:
		    return 0;
        case WM_KEYDOWN:
            switch ( wParam )
		    {
                case VK_ESCAPE:
			        PostQuitMessage(0);
					return 0;
                case VK_SPACE:
					g_bRender = !g_bRender;
			        return 0;
		    }
		    return 0;
		default:
		    return DefWindowProc( hWnd, message, wParam, lParam );
	}//End switch(..)
}// End WndProc(..)


/***************************************************************************/
/*                                                                         */
/* Name: RenderLoop(..)                                                    */
/* Date: 07-08-2004                                                        */
/* Orig: bkenwright@xbdev.net                                              */
/*                                                                         */
/* Desc: Gets called over and over again from our WinMain(..) function     */
/* when nothing is happening, so this always getting called, and this is   */
/* where we do all our rendering :) and opengl stuff.                      */
/* You could say this is the heart of our demo - as this is where all our  */
/* decisions are made.                                                     */
/* There is a small bit of win32 timming code at the start to limit the    */
/* maximum speed at which this gets called, as we want it to render and    */
/* animate our demo at the same speed on all the diffent computers whether */
/* its a AMD 400 or a P4 2.8Ghz.                                           */
/*                                                                         */
/***************************************************************************/

static void RenderLoop(HDC hDC, HWND hWnd)
{
	// statics allow us not to go using messy globals - so it stores the
	// value even after the function ends.
	static double OldTime = GetTickCount();
	double TimeNow = GetTickCount();
	// If less than 20milli seconds have occured since we last called our
	// function, then return without updating.  Else we save our new time
	// and continue to render.
	if( TimeNow - OldTime < 20 ) return;
	OldTime = TimeNow;

	// Used to rotate our torus donut around - so we can see it better
    static float theta = 0.0f;

	// Clear The Screen And The Depth Buffer
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	
			
	// Reset The View Matrix
    glLoadIdentity();

	// Create a perspective, so we can make things look smaller when they
	// get further away, and set a clipping area for our view.
    gluPerspective(30,      // fovy 
                   1.0,     // aspect
                   1.0,     // zNear
                   100.0);  // zFar
	
	// Set our camera position - so we are away from the object and can
	// see it more clearly
    gluLookAt(0, 0, -7,  // eyex, eyey, eyez          -> Position
              0, 0, 0,   // centerx, centery, centerz ->LookAt
              0, 1, 0);  // upx, upy, upz             ->Up

	// Rotate our object by a small amount.
    glRotatef( theta, 0.0f, 0.02f, 0.01f );

	// Create our 3D Donut shape - 20 rings, and 20 columns
    torus(20,20);
			
	// Flip our back buffer and show our shape.
    SwapBuffers( hDC );

	// Set the window text so the user has some information feedback
	char str[] = "Press Space to toggle wire frame (ESC Exit)";
	SetWindowText(hWnd, str);

	// Update the angle some more, so the next time round our object
	// will be rotated a bit more than last time.
    theta += 1.0f;
}//End RenderLoop(..)



/***************************************************************************/
/*                                                                         */
/* Name: StartOpenGL(..)                                                   */
/* Date: 07-08-2004                                                        */
/* Orig: bkenwright@xbdev.net                                              */
/*                                                                         */
/* Desc: With everything, there is settup code - and OpenGl is no          */
/* different.  So its been neatly placed here.  We need to setup or win32  */
/* window so that it uses opengl...and we set a few things, like if we are */
/* using double buffering, what type of pixel size we have etc.            */
/*                                                                         */
/***************************************************************************/

void StartOpenGL(HWND hWnd, HDC * hDC, HGLRC * hRC)
{
	PIXELFORMATDESCRIPTOR pfd;
	int format;
	
	// get the device context (DC)
	*hDC = GetDC( hWnd );
	
	// set the pixel format for the DC
	ZeroMemory( &pfd, sizeof( pfd ) );
	pfd.nSize = sizeof( pfd );
	pfd.nVersion = 1;
	pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = 24;
	pfd.cDepthBits = 16;
	pfd.iLayerType = PFD_MAIN_PLANE;
	format = ChoosePixelFormat( *hDC, &pfd );
	SetPixelFormat( *hDC, format, &pfd );
	
	// create and enable the render context (RC)
	*hRC = wglCreateContext( *hDC );
	wglMakeCurrent( *hDC, *hRC );

}//End StartOpenGL(..)


/***************************************************************************/
/*                                                                         */
/* Name: StartOpenGL(..)                                                   */
/* Date: 07-08-2004                                                        */
/* Orig: bkenwright@xbdev.net                                              */
/*                                                                         */
/* Desc: We always have to tidy up.  Can't go creating an OpenGL window    */
/* and not telling OpenGL that we've finished with the window and don't    */
/* need it anymore - which is what we do here.  We call a couple of the    */
/* opengl api's and tell it, that its we've finished with it.              */
/*                                                                         */
/***************************************************************************/

void EndOpenGL(HWND hWnd, HDC hDC, HGLRC hRC)
{
	wglMakeCurrent( NULL, NULL );
	wglDeleteContext( hRC );
	ReleaseDC( hWnd, hDC );
}//End EndOpenGL(..)


